import 'dart:async';
import 'dart:io';
import 'dart:ui' as ui;
import 'package:image/image.dart' as imgg;

import 'package:flutter/widgets.dart';
import 'package:xcommonwidgets/xcommonwidgets.dart';
import 'package:xfxwidgets/xMediaItem.dart';
import 'package:xtoolspackage/xtoolspackage.dart';

import './ImagePainter_ColorItem.dart';
import './ImagePainter_Controller.dart';
import './ImagePainter_RangeSlider.dart';
import './ImagePainter_SelectionItems.dart';
import './ImagePainter_TextDelegate.dart';
import './ImagePainter_TextDialog.dart';
import './ImagePaintr_Class.dart';
import 'package:flutter/material.dart';
import 'package:flutter/rendering.dart';
import 'package:flutter/services.dart';

///[ImagePainter] widget.
@immutable
class ImagePainter extends StatefulWidget {
  ImagePainter._({
    Key? key,
    this.height,
    this.width,
    this.placeHolder,
    this.isScalable,
    this.brushIcon,
    this.clearAllIcon,
    this.colorIcon,
    this.loadMedia,
    this.undoIcon,
    this.controlsAtTop = true,
    this.signatureBackgroundColor = Colors.white,
    this.colors,
    this.initialPaintMode,
    this.initialStrokeWidth,
    this.initialColor,
    this.onColorChanged,
    this.onStrokeWidthChanged,
    this.onPaintModeChanged,
    this.textDelegate,
    this.xMedia,
    this.backImage,
  }) : super(key: key);

  ///Constructor for loading image from network url.
  factory ImagePainter.network(
    String url, {
    required Key key,
    double? height,
    double? width,
    Widget? placeholderWidget,
    bool? scalable,
    List<Color>? colors,
    Widget? brushIcon,
    Widget? undoIcon,
    Widget? clearAllIcon,
    Widget? colorIcon,
    PaintMode? initialPaintMode,
    double? initialStrokeWidth,
    Color? initialColor,
    ValueChanged<PaintMode>? onPaintModeChanged,
    ValueChanged<Color>? onColorChanged,
    ValueChanged<double>? onStrokeWidthChanged,
    TextDelegate? textDelegate,
    bool? controlsAtTop,
  }) {
    return ImagePainter._(
      key: key,
      height: height,
      width: width,
      placeHolder: placeholderWidget,
      isScalable: scalable,
      colors: colors,
      brushIcon: brushIcon,
      undoIcon: undoIcon,
      colorIcon: colorIcon,
      clearAllIcon: clearAllIcon,
      initialPaintMode: initialPaintMode,
      initialColor: initialColor,
      initialStrokeWidth: initialStrokeWidth,
      onPaintModeChanged: onPaintModeChanged,
      onColorChanged: onColorChanged,
      onStrokeWidthChanged: onStrokeWidthChanged,
      textDelegate: textDelegate,
      controlsAtTop: controlsAtTop ?? true,
    );
  }

  ///Constructor for loading image from assetPath.

  ///Constructor for signature painting.
  factory ImagePainter.signature({
    required Key key,
    Color? signatureBgColor,
    double? height,
    double? width,
    List<Color>? colors,
    Widget? brushIcon,
    Widget? undoIcon,
    Widget? clearAllIcon,
    Widget? colorIcon,
    bool isNota = false,
    bool isNotaWithImage = false,
    ValueChanged<PaintMode>? onPaintModeChanged,
    ValueChanged<Color>? onColorChanged,
    ValueChanged<double>? onStrokeWidthChanged,
    TextDelegate? textDelegate,
    bool? controlsAtTop,
    Future Function(dynamic)? loadMedia,
    XMediaItem? xMedia,
  }) {
    return ImagePainter._(
      key: key,
      height: height,
      width: width,
      xMedia: xMedia,
      isScalable: false,
      colors: colors ?? List<Color>.empty(growable: true),
      loadMedia: loadMedia,
      signatureBackgroundColor: signatureBgColor ?? Colors.white,
      brushIcon: brushIcon,
      undoIcon: undoIcon,
      colorIcon: colorIcon,
      clearAllIcon: clearAllIcon,
      onPaintModeChanged: onPaintModeChanged,
      onColorChanged: onColorChanged,
      onStrokeWidthChanged: onStrokeWidthChanged,
      textDelegate: textDelegate,
      controlsAtTop: controlsAtTop ?? true,
    );
  }

  // ///Only accessible through [ImagePainter.network] constructor.
  // final String? networkUrl;

  // ///Only accessible through [ImagePainter.memory] constructor.
  // final Uint8List? byteArray;

  // ///Only accessible through [ImagePainter.file] constructor.
  // final File? file;

  // ///Only accessible through [ImagePainter.asset] constructor.
  // final String? assetPath;

  XMediaItem? xMedia;
  Image? backImage;

  ///Height of the Widget. Image is subjected to fit within the given height.
  double? height;

  ///Width of the widget. Image is subjected to fit within the given width.
  double? width;
  final Future Function(dynamic)? loadMedia;

  ///Widget to be shown during the conversion of provided image to [ui.Image].
  final Widget? placeHolder;

  ///Defines whether the widget should be scaled or not. Defaults to [false].
  final bool? isScalable;

  ///Signature mode background color
  final Color signatureBackgroundColor;

  ///List of colors for color selection
  ///If not provided, default colors are used.
  List<Color>? colors = List<Color>.empty(growable: true);

  ///Icon Widget of strokeWidth.
  final Widget? brushIcon;

  ///Widget of Color Icon in control bar.
  final Widget? colorIcon;

  ///Widget for Undo last action on control bar.
  final Widget? undoIcon;

  ///Widget for clearing all actions on control bar.
  final Widget? clearAllIcon;

  ///Define where the controls is located.
  ///`true` represents top.
  final bool controlsAtTop;

  ///Initial PaintMode.
  final PaintMode? initialPaintMode;

  //the initial stroke width
  final double? initialStrokeWidth;

  //the initial color
  final Color? initialColor;

  final ValueChanged<Color>? onColorChanged;

  final ValueChanged<double>? onStrokeWidthChanged;

  final ValueChanged<PaintMode>? onPaintModeChanged;

  //the text delegate
  final TextDelegate? textDelegate;

  @override
  ImagePainterState createState() => ImagePainterState();
}

///
class ImagePainterState extends State<ImagePainter> {
  final _repaintKey = GlobalKey();
  Image? _image;
  late Controller _controller;
  late final ValueNotifier<bool> _isLoaded;
  late final TextEditingController _textController;
  late final TransformationController _transformationController_Payload;
  late final TransformationController _transformationController_Overlay;
  final _imagePainterState_Key = GlobalKey<ImagePainterState>();

  int _strokeMultiplier = 1;
  late TextDelegate textDelegate;

  bool modifiedActive = false;
  bool isBusy = true;

  @override
  void initState() {
    super.initState();
    if (widget.colors == null) {
      widget.colors!.addAll([Colors.yellow, Colors.black, Colors.red, Colors.white, Colors.purple, Colors.blue, Colors.green]);
    }
    _isLoaded = ValueNotifier<bool>(false);
    _controller = Controller();

    loadMediaAndImage();
  }

  Uint8List createTransparentImage_fromPayLoad(Uint8List payloadOri) {
    final imgOri = imgg.decodeJpg(payloadOri)!;
    // img.decodeJpg(payloadOri)!;

    //final image = img.Image(width: imgOri.width, height: imgOri.height, backgroundColor: img.ColorFloat32(0x00000000));
    // final image = img.Image(width: imgOri.width, height: imgOri.height, numChannels: 4, format: img.Format.uint8);
    final imgg.Image image = imgg.Image(width: imgOri.width, height: imgOri.height, numChannels: 4, format: imgg.Format.uint8);

    // Imposta tutti i pixel come completamente trasparenti (alpha = 0)
    // Converti l'immagine in Uint8List
    //final pngBytes = image.toUint8List(); // Uint8List.fromList(img.encodePng(image));
    //var image = img.Image.from(imgOri, noPixels: true);
    image.clear(imgg.ColorUint8.rgba(0x00, 0x00, 0x00, 0x00));
    final pngBytes = imgg.encodePng(image);
    return pngBytes;
  }

  Image createTransparentImage(Size size) {
    final imgg.Image image = imgg.Image(
      width: size.width.toInt(),
      height: size.height.toInt(),
      numChannels: 4,
      format: imgg.Format.uint4, //uint8
    );
    image.clear(imgg.ColorUint4.rgba(0x00, 0x00, 0x00, 0x00));
    return Image.memory(image.getBytes());
  }

  Future<Image> createTransparentUiImage(int width, int height) async {
    final imgg.Image imggi = imgg.Image(
      width: width,
      height: height,
      numChannels: 4,
      format: imgg.Format.uint4, //uint8
    );
    imggi.clear(imgg.ColorUint4.rgba(0x00, 0x00, 0x00, 0x00));
//Image imagepr = Image.memory(imggi.getBytes());
    //  var ib = await ui.ImmutableBuffer.fromUint8List(imggi.buffer.asUint8List());
    //final id = await ui.ImageDescriptor.encoded(ib);
    var r = Image.memory(imggi.getBytes());

    // var id = ui.ImageDescriptor.raw(ui.ImmutableBuffer(image. buffer), width: image.width, height: image.height, pixelFormat: ui.PixelFormat.rgba8888);

    // var r = ui.Image(id);
    return r;
  }

  loadMediaAndImage() async {
    //carichiamo i payloads se uno è vuoto
    if (widget.xMedia!.payload == null) {
      await widget.loadMedia!(widget.xMedia);
    } else {}

    if (widget.xMedia!.payload != null) {
      widget.backImage = Image.memory(
        widget.xMedia!.payload!,
        // height: MediaQuery.of(context).size.height / 2,
        // width: MediaQuery.of(context).size.width / 2,
      );
    }

    if (widget.xMedia!.overlaySurface != null) {
      _image = await _convertImage(widget.xMedia!.overlaySurface!);
      if (_image == null) {
        throw ("Image couldn't be resolved from xMedia!.overlaySurface.");
      } else {
        isBusy = false;
        _setStrokeMultiplier();
      }
    } else {
      _image = await createTransparentUiImage(_image!.width!.toInt(), _image!.height!.toInt());
      isBusy = false;
      _setStrokeMultiplier();
      _isLoaded.value = true;
    }

    _controller.naturalSize = imageSize;
    //if (widget.isSignature) {
    _controller.update(
      mode: PaintMode.none,
      color: Colors.black,
    );
    // } else {
    //   _controller.update(
    //     mode: widget.initialPaintMode,
    //     strokeWidth: widget.initialStrokeWidth,
    //     color: widget.initialColor,
    //   );
    // }
    _textController = TextEditingController();
    _transformationController_Payload = TransformationController();
    _transformationController_Overlay = TransformationController();
    textDelegate = widget.textDelegate ?? TextDelegate();
    _isLoaded.value = true;
  }

  @override
  void dispose() {
    _controller.dispose();
    _isLoaded.dispose();
    _textController.dispose();
    _transformationController_Payload.dispose();
    _transformationController_Overlay.dispose();
    super.dispose();
  }

  bool get isEdited => _controller.paintHistory.isNotEmpty;

  Size get imageSize => Size(_image == null ? widget.width ?? double.maxFinite : _image?.width!.toDouble() ?? 0, _image == null ? widget.height ?? double.maxFinite : _image?.height!.toDouble() ?? 0);

  ///Dynamically sets stroke multiplier on the basis of widget size.
  ///Implemented to avoid thin stroke on high res images.
  _setStrokeMultiplier() {
    if ((_image!.height! + _image!.width!) > 1000) {
      _strokeMultiplier = (_image!.height! + _image!.width!) ~/ 1000;
    }
    _controller.update(strokeMultiplier: _strokeMultiplier);
  }

  ///Completer function to convert asset or file image to [ui.Image] before drawing on custompainter.
  Future<Image> _convertImage(Uint8List bytes) async {
    Image.memory(bytes);
    return Image.memory(bytes);
    // final completer = Completer<Image>();
    // ui.decodeImageFromList(img, (image) {
    //   return completer.complete(image);
    // });
    // return completer.future;
  }

  ///Completer function to convert network image to [ui.Image] before drawing on custompainter.
  Future<ui.Image> _loadNetworkImage(String path) async {
    final completer = Completer<ImageInfo>();
    final img = NetworkImage(path);
    img.resolve(const ImageConfiguration()).addListener(ImageStreamListener((info, _) => completer.complete(info)));
    final imageInfo = await completer.future;
    _isLoaded.value = true;
    return imageInfo.image;
  }

  @override
  Widget build(BuildContext context) {
    return ValueListenableBuilder<bool>(
      valueListenable: _isLoaded,
      builder: (_, loaded, __) {
        if (loaded) {
          return _paintSignature();
          //return widget.isSignature ? _paintSignature() : _paintImage();
        } else {
          return Container(
            height: widget.height ?? double.maxFinite,
            width: widget.width ?? double.maxFinite,
            child: Center(
              child: widget.placeHolder ?? const CircularProgressIndicator(),
            ),
          );
        }
      },
    );
  }

  Widget _paintSignature() {
    return Container(
        height: MediaQuery.of(context).size.height / 1.2,
        child: StatefulBuilder(builder: (context, setState) {
          return XAlertDialog(
              child: isBusy
                  ? XStatefulWidgetBase.LoaderIndicator(widget.xMedia!.isSel)
                  : Container(
                      color: Colors.white,
                      height: widget.height ?? double.maxFinite,
                      width: widget.width ?? double.maxFinite,
                      child: Column(children: [
                        modifiedActive
                            ? widget.controlsAtTop
                                ? _buildControls()
                                : Container()
                            : Container(),
                        Expanded(
                            child: FittedBox(
                                alignment: FractionalOffset.center,
                                child: RepaintBoundary(
                                    key: _repaintKey,
                                    child: ClipRect(
                                        child: AnimatedBuilder(
                                            animation: _controller,
                                            builder: (context, child) {
                                              return InteractiveViewer(
                                                  transformationController: _transformationController_Overlay,
                                                  maxScale: 3,
                                                  minScale: 0.2,
                                                  panEnabled: _controller.mode == PaintMode.none,
                                                  scaleEnabled: _controller.mode == PaintMode.none,
                                                  onInteractionStart: _scaleStartGesture,
                                                  onInteractionUpdate: _scaleUpdateGesture,
                                                  onInteractionEnd: _scaleEndGesture,
                                                  child: Stack(children: [
                                                    widget.xMedia!.payload != null
                                                        ? Container(
                                                            //margin: EdgeInsets.only(top: 20, left: 10),
                                                            //////child: Image.memory(widget.xMedia!.payload!,
                                                            child: widget.backImage,
                                                          )
                                                        : Container(),
                                                    CustomPaint(
                                                        size: Size(_controller.naturalSize.width, _controller.naturalSize.height),
                                                        willChange: true,
                                                        isComplex: true,
                                                        painter: DrawImage(
                                                          image: _image,
                                                          backgroundColor: Colors.transparent,
                                                          isSignature: true,
                                                          controller: _controller,
                                                        )),
                                                  ]));
                                            }))))),
                        modifiedActive
                            ? widget.controlsAtTop.not()
                                ? _buildControls()
                                : Container()
                            : Container(),
                        SizedBox(height: MediaQuery.of(context).padding.bottom)
                      ])),
              content_insetPadding: EdgeInsets.symmetric(horizontal: 4),
              actionsBTNarea_Padding: EdgeInsets.all(0),
              actionsBTNarea_Child: Row(children: [
                Expanded(
                    child: XBtnbase(
                  height: 40,
                  backGroundColor: Colors.transparent,
                  onPressed: () {
                    modifiedActive = false;
                    widget.xMedia!.isModified = false;
                    Navigator.pop(context, false);
                  },
                  elevation: 0,
                  label: "Chiudi",
                  label_Style: TextStyle(color: Colors.red, fontSize: 18),
                )),
                modifiedActive.not()
                    ? Expanded(
                        child: XBtnbase(
                        height: 40,
                        width: 120,
                        backGroundColor: Colors.transparent,
                        onPressed: () {
                          setState(() {
                            modifiedActive = true;
                            _controller.setMode(PaintMode.freeStyle);
                          });
                        },
                        elevation: 0,
                        label: "Annota",
                        label_Style: TextStyle(color: Colors.orange[600], fontSize: 18),
                      ))
                    : Container(),
                widget.xMedia!.isModified
                    ? Expanded(
                        child: XBtnbase(
                        height: 40,
                        backGroundColor: Colors.transparent,
                        onPressed: () async {
                          modifiedActive = false;
                          setState(() {
                            isBusy = true;
                          });
                          await cmd_Save();
                          setState(() {
                            isBusy = false;
                            widget.xMedia!.isModified = false;
                          });
                          Navigator.pop(context, true);
                        },
                        elevation: 0,
                        label: "Salva",
                        label_Style: TextStyle(color: Colors.green, fontSize: 18),
                      ))
                    : Container()
              ]));
        }));
  }

  cmd_Save() async {
    if (_image != null) {
      widget.xMedia!.overlaySurface = await _imagePainterState_Key.currentState!.exportImage();
    }
    print(widget.xMedia!.overlaySurface!.length);
    if (widget.xMedia!.miniatura == null) {
      var x = createThumbnails_fromImage(widget.xMedia!.payload!);
      widget.xMedia!.miniatura = createThumbnailsWithOverlay(widget.xMedia!.overlaySurface!, x);
    } else {
      widget.xMedia!.miniatura = createThumbnailsWithOverlay(widget.xMedia!.overlaySurface!, widget.xMedia!.miniatura!);
    }
  }

  Uint8List createThumbnails_fromImage(Uint8List image_payload, {int width = 80, int height = 100}) {
    final imgOri = imgg.decodeJpg(image_payload)!;
    var resizedImg = imgg.copyResize(imgOri, width: width, height: height);
    final pngBytes = imgg.encodePng(resizedImg);
    return pngBytes;
  }

  Uint8List createThumbnailsWithOverlay(Uint8List overlay_payload, Uint8List image_Thumbnails, {int width = 80, int height = 100}) {
    final overLayImg = imgg.decodePng(overlay_payload)!;
    var resizedOverlayImg = imgg.Image.fromResized(overLayImg, width: width, height: height);
    var imgThumb = imgg.decodePng(image_Thumbnails)!;

    var newMiniature = imgg.compositeImage(imgThumb, resizedOverlayImg);

    final pngBytes = imgg.encodePng(newMiniature);
    return pngBytes;
  }

  _scaleStartGesture(ScaleStartDetails onStart) {
    if (_controller.mode == PaintMode.none) return;

    final _zoomAdjustedOffset = _transformationController_Overlay.toScene(onStart.localFocalPoint);
    _controller.setStart(_zoomAdjustedOffset);
    _controller.addOffsets(_zoomAdjustedOffset);
  }

  ///Fires while user is interacting with the screen to record painting.
  void _scaleUpdateGesture(ScaleUpdateDetails onUpdate) {
    if (_controller.mode == PaintMode.none) return;
    if (widget.xMedia!.isModified.not()) {
      setState(() {
        widget.xMedia!.isModified = true;
      });
    }
    final _zoomAdjustedOffset = _transformationController_Overlay.toScene(onUpdate.localFocalPoint);
    _controller.setInProgress(true);
    if (_controller.start == null) {
      _controller.setStart(_zoomAdjustedOffset);
    }
    _controller.setEnd(_zoomAdjustedOffset);
    if (_controller.mode == PaintMode.freeStyle) {
      _controller.addOffsets(_zoomAdjustedOffset);
    }
    if (_controller.onTextUpdateMode) {
      _controller.paintHistory.lastWhere((element) => element.mode == PaintMode.text).offsets = [_zoomAdjustedOffset];
    }
  }

  ///Fires when user stops interacting with the screen.
  void _scaleEndGesture(ScaleEndDetails onEnd) {
    if (_controller.mode == PaintMode.none) return;

    _controller.setInProgress(false);
    if (_controller.start != null && _controller.end != null && (_controller.mode == PaintMode.freeStyle)) {
      _controller.addOffsets(null);
      _addFreeStylePoints();
      _controller.offsets.clear();
    } else if (_controller.start != null && _controller.end != null && _controller.mode != PaintMode.text) {
      _addEndPoints();
    }
    _controller.resetStartAndEnd();
  }

  void _addEndPoints() => _addPaintHistory(
        PaintInfo(
          offsets: <Offset?>[_controller.start, _controller.end],
          mode: _controller.mode,
          color: _controller.color,
          strokeWidth: _controller.scaledStrokeWidth,
          fill: _controller.fill,
        ),
      );

  void _addFreeStylePoints() => _addPaintHistory(
        PaintInfo(
          offsets: <Offset?>[..._controller.offsets],
          mode: PaintMode.freeStyle,
          color: _controller.color,
          strokeWidth: _controller.scaledStrokeWidth,
        ),
      );

  ///Provides [ui.Image] of the recorded canvas to perform action.
  Future<ui.Image> _renderImage() async {
    final recorder = ui.PictureRecorder();
    final canvas = Canvas(recorder);
    final painter = DrawImage(image: _image, controller: _controller);
    final size = Size(_image!.width!.toDouble(), _image!.height!.toDouble());
    painter.paint(canvas, size);
    return recorder.endRecording().toImage(size.width.floor(), size.height.floor());
  }

  PopupMenuItem _showOptionsRow() {
    return PopupMenuItem(
      enabled: false,
      child: Center(
        child: SizedBox(
          child: Wrap(
            children: paintModes(textDelegate)
                .map(
                  (item) => SelectionItems(
                    data: item,
                    isSelected: _controller.mode == item.mode,
                    onTap: () {
                      if (widget.onPaintModeChanged != null) {
                        widget.onPaintModeChanged!(item.mode);
                      }
                      _controller.setMode(item.mode);

                      Navigator.of(context).pop();
                      if (item.mode == PaintMode.text) {
                        _openTextDialog();
                      }
                    },
                  ),
                )
                .toList(),
          ),
        ),
      ),
    );
  }

  PopupMenuItem _showRangeSlider() {
    return PopupMenuItem(
      enabled: false,
      child: SizedBox(
        width: double.maxFinite,
        child: AnimatedBuilder(
          animation: _controller,
          builder: (_, __) {
            return RangedSlider(
              value: _controller.strokeWidth,
              onChanged: (value) {
                _controller.setStrokeWidth(value);
                if (widget.onStrokeWidthChanged != null) {
                  widget.onStrokeWidthChanged!(value);
                }
              },
            );
          },
        ),
      ),
    );
  }

  PopupMenuItem _showColorPicker() {
    return PopupMenuItem(
      enabled: false,
      child: Center(
        child: Wrap(
          alignment: WrapAlignment.center,
          spacing: 10,
          runSpacing: 10,
          children: (widget.colors ?? editorColors).map((color) {
            return ColorItem(
              isSelected: color == _controller.color,
              color: color,
              onTap: () {
                _controller.setColor(color);
                if (widget.onColorChanged != null) {
                  widget.onColorChanged!(color);
                }
                Navigator.pop(context);
              },
            );
          }).toList(),
        ),
      ),
    );
  }

  ///Generates [Uint8List] of the [ui.Image] generated by the [renderImage()] method.
  ///Can be converted to image file by writing as bytes.
  Future<Uint8List?> exportImage() async {
    late ui.Image _convertedImage;
    //  questo fa un mix della foto in back + l'overlay
    // final _boundary = _repaintKey.currentContext!.findRenderObject() as RenderRepaintBoundary;
    // _convertedImage = await _boundary.toImage(pixelRatio: 3);

    //a noi interessa SOLO _image (ns overlay puro)
    _convertedImage = await _renderImage();
    final byteData = await _convertedImage.toByteData(format: ui.ImageByteFormat.png);
    return byteData?.buffer.asUint8List();
  }

  void _addPaintHistory(PaintInfo info) {
    if (info.mode != PaintMode.none) {
      _controller.addPaintInfo(info);
    }
  }

  void _openTextDialog() {
    _controller.setMode(PaintMode.text);
    final fontSize = 6 * _controller.strokeWidth;
    TextDialog.show(
      context,
      _textController,
      fontSize,
      _controller.color,
      textDelegate,
      onFinished: (context) {
        if (_textController.text.isNotEmpty) {
          _addPaintHistory(
            PaintInfo(
              mode: PaintMode.text,
              text: _textController.text,
              offsets: [],
              color: _controller.color,
              strokeWidth: _controller.scaledStrokeWidth,
            ),
          );
          _textController.clear();
        }
        Navigator.of(context).pop();
      },
    );
  }

  Widget _buildControls() {
    return Container(
      padding: const EdgeInsets.all(4),
      color: Colors.grey[200],
      child: Row(
        children: [
          AnimatedBuilder(
            animation: _controller,
            builder: (_, __) {
              final icon = paintModes(textDelegate).firstWhere((item) => item.mode == _controller.mode).icon;
              return PopupMenuButton(
                tooltip: textDelegate.changeMode,
                shape: ContinuousRectangleBorder(borderRadius: BorderRadius.circular(40)),
                icon: Icon(icon, color: Colors.black),
                itemBuilder: (_) => [_showOptionsRow()],
              );
            },
          ),
          AnimatedBuilder(
            animation: _controller,
            builder: (_, __) {
              return PopupMenuButton(
                padding: const EdgeInsets.symmetric(vertical: 10),
                shape: ContinuousRectangleBorder(borderRadius: BorderRadius.circular(20)),
                tooltip: textDelegate.changeColor,
                icon: widget.colorIcon ??
                    Container(
                      padding: const EdgeInsets.all(2.0),
                      decoration: BoxDecoration(
                        shape: BoxShape.circle,
                        border: Border.all(color: Colors.black),
                        color: _controller.color,
                      ),
                    ),
                itemBuilder: (_) => [_showColorPicker()],
              );
            },
          ),
          PopupMenuButton(
            tooltip: textDelegate.changeBrushSize,
            shape: ContinuousRectangleBorder(borderRadius: BorderRadius.circular(20)),
            icon: widget.brushIcon ?? Icon(Icons.brush, color: Colors.black),
            itemBuilder: (_) => [_showRangeSlider()],
          ),
          // AnimatedBuilder(
          //   animation: _controller,
          //   builder: (_, __) {
          //     if (_controller.canFill()) {
          //       return Row(children: [
          //         Checkbox.adaptive(
          //             value: _controller.shouldFill,
          //             checkColor: Colors.black,
          //             onChanged: (val) {
          //               _controller.update(fill: val);
          //             }),
          //         Text('Riempi', style: TextStyle(color: Colors.black))
          //       ]);
          //     } else {
          //       return const SizedBox();
          //     }
          //   },
          // ),
          const Spacer(),
          IconButton(
            tooltip: textDelegate.undo,
            icon: widget.undoIcon ?? Icon(Icons.reply, color: Colors.black),
            onPressed: () => _controller.undo(),
          ),
          IconButton(
            tooltip: textDelegate.clearAllProgress,
            icon: widget.clearAllIcon ?? Icon(Icons.clear, color: Colors.black),
            onPressed: () => _controller.clear(),
          ),
          IconButton(
            tooltip: "Premi per avere più carta",
            icon: Icon(Icons.image_aspect_ratio, color: Colors.black),
            onPressed: () => _controller.morePaper(),
          ),
        ],
      ),
    );
  }
}

// // ///paints image on given constrains for drawing if image is not null.
// // Widget _paintImage() {
// //   return Container(
// //     height: widget.height ?? double.maxFinite,
// //     width: widget.width ?? double.maxFinite,
// //     child: Column(
// //       children: [
// //         if (widget.controlsAtTop) _buildControls(),
// //         Expanded(
// //           child: FittedBox(
// //             alignment: FractionalOffset.center,
// //             child: ClipRect(
// //               child: AnimatedBuilder(
// //                 animation: _controller,
// //                 builder: (context, child) {
// //                   return InteractiveViewer(
// //                     transformationController: _transformationController_Overlay,
// //                     maxScale: 2.4,
// //                     minScale: 1,
// //                     panEnabled: _controller.mode == PaintMode.none,
// //                     scaleEnabled: widget.isScalable!,
// //                     onInteractionUpdate: _scaleUpdateGesture,
// //                     onInteractionEnd: _scaleEndGesture,
// //                     child: CustomPaint(
// //                       size: imageSize,
// //                       willChange: true,
// //                       isComplex: true,
// //                       painter: DrawImage(
// //                         image: _image,
// //                         controller: _controller,
// //                       ),
// //                     ),
// //                   );
// //                 },
// //               ),
// //             ),
// //           ),
// //         ),
// //         if (!widget.controlsAtTop) _buildControls(),
// //         SizedBox(height: MediaQuery.of(context).padding.bottom)
// //       ],
// //     ),
// //   );
// // }
